---@param name string
---@param speed number
---@return Fielder
function newFielder(name, speed)
    return {
        name = name,
        speed = speed,
    }
end

-- selene: allow(unscoped_variables)
Field = {
    fielders = {
        first = newFielder("First", 40),
        second = newFielder("Second", 40),
        shortstop = newFielder("Shortstop", 40),
        third = newFielder("Third", 40),
        pitcher = newFielder("Pitcher", 30),
        catcher = newFielder("Catcher", 35),
        left = newFielder("Left", 40),
        center = newFielder("C.Center", 40),
        right = newFielder("Right", 40),
    },
}

--- Actually only benches the infield, because outfielders are far away!
---@param position XYPair
function Field.benchTo(self, position)
    self.fielders.first.target = position
    self.fielders.second.target = position
    self.fielders.shortstop.target = position
    self.fielders.third.target = position
    self.fielders.pitcher.target = position
    self.fielders.catcher.target = position
end

--- Resets the target positions of all fielders to their defaults (at their field positions).
---@param fromOffTheField XYPair | nil If provided, also sets all runners' current position to one centralized location.
function Field.resetFielderPositions(self, fromOffTheField)
    if fromOffTheField then
        for _, fielder in pairs(self.fielders) do
            fielder.x = fromOffTheField.x
            fielder.y = fromOffTheField.y
        end
    end

    self.fielders.first.target = utils.xy(C.Screen.W - 65, C.Screen.H * 0.48)
    self.fielders.second.target = utils.xy(C.Screen.W * 0.70, C.Screen.H * 0.30)
    self.fielders.shortstop.target = utils.xy(C.Screen.W * 0.30, C.Screen.H * 0.30)
    self.fielders.third.target = utils.xy(C.Screen.W * 0.1, C.Screen.H * 0.48)
    self.fielders.pitcher.target = utils.xy(C.PitcherStartPos.x, C.PitcherStartPos.y)
    self.fielders.catcher.target = utils.xy(C.Screen.W * 0.475, C.Screen.H * 0.92)
    self.fielders.left.target = utils.xy(C.Screen.W * -1, C.Screen.H * -0.2)
    self.fielders.center.target = utils.xy(C.Center.x, C.Screen.H * -0.4)
    self.fielders.right.target = utils.xy(C.Screen.W * 2, self.fielders.left.target.y)
end

---@param fielder Fielder
---@param ballPos XYPair
---@return boolean isTouchingBall
local function updateFielderPosition(deltaSeconds, fielder, ballPos)
    if fielder.target ~= nil then
        if not utils.moveAtSpeed(fielder, fielder.speed * deltaSeconds, fielder.target) then
            fielder.target = nil
        end
    end

    return utils.distanceBetweenPoints(fielder, ballPos) < C.BallCatchHitbox
end

--- Selects the nearest fielder to move toward the given coordinates.
--- Other fielders should attempt to cover their bases
function Field.haveSomeoneChase(self, ballDestX, ballDestY)
    local chasingFielder = utils.getNearestOf(self.fielders, ballDestX, ballDestY)
    chasingFielder.target = { x = ballDestX, y = ballDestY }

    for _, base in ipairs(C.Bases) do
        local nearest = utils.getNearestOf(self.fielders, base.x, base.y, function(fielder)
            -- For now, skipping the pitcher because they're considered closer to 2B than second or shortstop
            return fielder ~= chasingFielder and fielder ~= self.fielders.pitcher
        end)
        nearest.target = base
    end
end

---@param ball XYPair
---@return Fielder | nil fielderTouchingBall nil if no fielder is currently touching the ball
function Field.updateFielderPositions(self, ball, deltaSeconds)
    local fielderTouchingBall
    for _, fielder in pairs(self.fielders) do
        local isTouchingBall = updateFielderPosition(deltaSeconds, fielder, ball)
        if isTouchingBall then
            fielderTouchingBall = fielder
        end
    end
    return fielderTouchingBall
end